/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <boost/lambda/lambda.hpp>
#include <omp.h>

#include <gsl/gsl_sort.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_deriv.h>

#include <iostream>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <numeric>
#include <ext/algorithm>

#include "AguiarGopinath.hh"
#include "hp_filter.hh"

AguiarGopinath::AguiarGopinath(model_t model_arg, double gam_arg, double beta_arg,
                               double lambda_arg, double delta_arg, double convergence_tol_arg, size_t ngrid_a_arg, size_t ngrid_y_arg)
  : model(model_arg),
    gam(gam_arg), r(0.01), beta(beta_arg), lambda(lambda_arg),
    delta(delta_arg), mu_g(1.006),
    sigma_z(model == Model1 ? 0.034 : 0.0),
    mu_z(-0.5*gsl_pow_2(sigma_z)),
    rho_z(model == Model1 ? 0.9 : NAN),
    sigma_g(model == Model1 ? 0 : 0.03),
    rho_g(model == Model1 ? NAN : 0.17),
    mu_y(log(mu_g)-0.5*gsl_pow_2(sigma_g)/(1-gsl_pow_2(rho_g))),
    min_a(model == Model1 ? -0.45 : -0.3), max_a(0.0),
    min_y(model == Model1 ? mu_z - 6*sigma_z/sqrt(1-gsl_pow_2(rho_z)) : mu_y - 6*sigma_g/sqrt(1-gsl_pow_2(rho_g))),
    max_y(model == Model1 ? mu_z + 6*sigma_z/sqrt(1-gsl_pow_2(rho_z)) : mu_y + 6*sigma_g/sqrt(1-gsl_pow_2(rho_g))),
    ngrid_a(ngrid_a_arg), ngrid_y(ngrid_y_arg), ngrid_a1(ngrid_a_arg),
    width(6),
    VBint(ngrid_y, gsl_interp_cspline),
    gauss(model == Model1 ? sigma_z : sigma_g, 16, GaussQuadrature::Legendre),
    convergence_tol(convergence_tol_arg),
    root_tol(1e-6),
    has_converged_flag(false)
{
  grid_y = new double[ngrid_y];
  for (size_t i = 0; i < ngrid_y; i++)
    grid_y[i] = min_y+((double) i)/((double) (ngrid_y-1))*(max_y-min_y);

  grid_a1 = new double[ngrid_a1];
  for (size_t i = 0; i < ngrid_a1; i++)
    grid_a1[i] = min_a+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a);
}

AguiarGopinath::AguiarGopinath(const AguiarGopinath &arg)
  : model(arg.model), gam(arg.gam), r(arg.r), beta(arg.beta), lambda(arg.lambda),
    delta(arg.delta), mu_g(arg.mu_g), sigma_z(arg.sigma_z), mu_z(arg.mu_z),
    rho_z(arg.rho_z), sigma_g(arg.sigma_g), rho_g(arg.rho_g), mu_y(arg.mu_y),
    min_a(arg.min_a), max_a(arg.max_a), min_y(arg.min_y), max_y(arg.max_y),
    ngrid_a(arg.ngrid_a), ngrid_y(arg.ngrid_y), ngrid_a1(arg.ngrid_a1),
    width(arg.width),
    VBint(arg.VBint),
    gauss(arg.gauss),
    convergence_tol(arg.convergence_tol),
    root_tol(arg.root_tol),
    has_converged_flag(arg.has_converged_flag)
{
  grid_y = new double[ngrid_y];
  __gnu_cxx::copy_n(arg.grid_y, ngrid_y, grid_y);
  grid_a1 = new double[ngrid_a1];
  __gnu_cxx::copy_n(arg.grid_a1, ngrid_a1, grid_a1);
}

AguiarGopinath::~AguiarGopinath()
{
  delete[] grid_y;
  delete[] grid_a1;
}

double
AguiarGopinath::q(double y, double a1, double *deriv) const
{
  assert(!std::isnan(y) && !std::isnan(a1));

  if (a1 >= 0)
    {
      if (deriv != NULL)
        *deriv = 0;
      
      return 1.0/(1.0+r);
    }
  
  try
    {
      double y1_threshold = default_threshold(y, a1);

      double cond_mean, Erepay1;
      switch(model)
        {
        case Model1:
          cond_mean = (1-rho_z)*mu_z + rho_z*y;
          Erepay1 = gsl_cdf_gaussian_Q(y1_threshold - cond_mean, sigma_z);
          break;
        case Model2:
          cond_mean = (1-rho_g)*mu_y + rho_g*y;
          Erepay1 = gsl_cdf_gaussian_Q(y1_threshold - cond_mean, sigma_g);
          break;
        }

      if (deriv != NULL)
        {
          const double epsdiff = 1e-8;

          gsl_function F;
          double abserr, VG_deriv_a;
          VG_deriv_params par = { y1_threshold, this };
          F.function = &AguiarGopinath::VG_deriv;
          F.params = &par;
          gsl_deriv_central(&F, a1, epsdiff, &VG_deriv_a, &abserr);

          difference_params params = { a1, this };

          F.function = &AguiarGopinath::difference;
          F.params = &params;
          double difference_diff;
          gsl_deriv_central(&F, y1_threshold, epsdiff, &difference_diff, &abserr);

          if (std::isfinite(VG_deriv_a) && std::isfinite(difference_diff))
            *deriv = gsl_ran_gaussian_pdf(y1_threshold - cond_mean, sigma_g)/(1+r)
              *VG_deriv_a/difference_diff;
          else
            *deriv = NAN;
        }
      

      return Erepay1/(1.0+r);
    }
  catch (LowerBound &e)
    {
      if (deriv != NULL)
        *deriv = 0;
      return 1.0/(1.0+r);
    }
  catch (UpperBound &e)
    {
      if (deriv != NULL)
        *deriv = 0;
      return 0.0;
    }
}

void
AguiarGopinath::quadrature_points_y(double y, double *quad_points) const
{
  switch (model)
    {
    case Model1:
      std::transform(gauss.get_points(), gauss.get_points() + gauss.npoints, quad_points,
                     boost::lambda::_1+(1-rho_z)*mu_z+rho_z*y);
      break;
    case Model2:
      std::transform(gauss.get_points(), gauss.get_points() + gauss.npoints, quad_points,
                     boost::lambda::_1+(1-rho_g)*mu_y + rho_g*y);
      break;
    }
}


double
AguiarGopinath::spread(double y, double a1) const
{
  assert(!std::isnan(y) && !std::isnan(a1));

  return gsl_pow_4(1/q(y, a1)/(1.0+r)) - 1.0;
}

double
AguiarGopinath::cons(double a, double y, state_t decision) const
{
  assert(!std::isnan(y) && !std::isnan(a));

  double g = (model == Model1 ? mu_g : exp(y));
  double a1;
  switch (decision)
    {
    case Default:
    case WaitRedemption:
      return (1-delta)*gdp(y);
    case Redemption:
    case Repay:
      a1 = a1_dr(a, y);
      return gdp(y)+a-g*a1*q(y, a1);
    }
}

void
AguiarGopinath::simulate(double *y, double *a, state_t *d, size_t n, gsl_rng *rng) const
{
  for (size_t i = 0; i < n-1; i++)
    {
      if (model == Model1)
        y[i+1] = (1-rho_z)*mu_z+rho_z*y[i]+gsl_ran_gaussian(rng, sigma_z);
      else
        y[i+1] = (1-rho_g)*mu_y+rho_g*y[i]+gsl_ran_gaussian(rng, sigma_g);

      if (d[i] == Default || d[i] == WaitRedemption)
        {
          if (gsl_rng_uniform(rng) < lambda)
            {
              d[i+1] = Redemption;
              a[i+1] = a1_dr(0.0, y[i]);
            }
          else
            {
              d[i+1] = WaitRedemption;
              a[i+1] = 0.0;
            }
        }
      else
        {
          if (def(a[i], y[i]))
            {
              d[i+1] = Default;
              a[i+1] = 0.0;
            }
          else
            {
              d[i+1] = Repay;
              a[i+1] = a1_dr(a[i], y[i]);
            }
        }
    }
}

void
AguiarGopinath::print_simulation(const double *y, const double *a, const state_t *d, size_t n) const
{
  for (size_t i = 0; i < n-1; i++)
    {
      std::cout << "State: y=" << gdp(y[i]) << ", a=" << a[i] << ", credit history=";
      switch (d[i])
        {
        case Default:
        case WaitRedemption:
          std::cout << "bad";
          break;
        case Repay:
        case Redemption:
          std::cout << "good";
          break;
        }
      std::cout << "; Decision: c=" << cons(a[i], y[i], d[i+1]) << ", ";
      switch (d[i+1])
        {
        case Default:
          std::cout << "default, VB=" << VB(y[i]);
          break;
        case WaitRedemption:
          std::cout << "wait redemption, VB=" << VB(y[i]);
          break;
        case Repay:
          std::cout << "repay, a1=" << a[i+1] << ", rs=" << spread(y[i], a[i+1]) << ", VG=" << VG(a[i], y[i]);
          break;
        case Redemption:
          std::cout << "redemption, a1=" << a[i+1] << ", rs=" << spread(y[i], a[i+1]) << ", VG=" << VG(0.0, y[i]);
          break;
        }
      std::cout << std::endl;
    }
}

void
AguiarGopinath::csv_output(const double *y, const double *a, const state_t *d, size_t n, const std::string &filename) const
{
  std::ofstream output(filename.c_str());

  output << std::scientific << std::setprecision(10);

  output << "y,a,c,d,a1,rs,V" << std::endl;
  for (size_t i = 0; i < n-1; i++)
    {
      output << gdp(y[i]) << "," << a[i] << "," << cons(a[i], y[i], d[i+1]) << ",";

      switch (d[i+1])
        {
        case Default:
          output << "default,nan,0," << VB(y[i]);
          break;
        case WaitRedemption:
          output << "wait_redemption,nan,0," << VB(y[i]);
          break;
        case Repay:
          output << "repay," << a[i+1] << "," << spread(y[i], a[i+1]) << "," << VG(a[i], y[i]);
          break;
        case Redemption:
          output << "redemption," << a[i+1] << "," << spread(y[i], a[i+1]) << "," << VG(0.0, y[i]);
          break;
        }
      output << std::endl;
    }

  output.close();
}

void
AguiarGopinath::simulation_statistics(const double *y, const double *a, const state_t *d, size_t n, size_t warmup, statistics_t &r, double scale) const
{
  assert(n > warmup+1);

  double *gdp_x = new double[n-warmup-1];
  double *cons_x = new double[n-warmup-1];
  double *tb_y_x = new double[n-warmup-1];
  double *spread_x = new double[n-warmup-1];
  double *debt_y_x = new double[n-warmup-1];
  size_t ndefaults = 0;

  double trend = mu_g;

  for (size_t i = warmup; i < n-1; i++)
    {
      gdp_x[i-warmup] = log(gdp(y[i]) * trend);

      if (d[i+1] == Repay || d[i+1] == Redemption)
        {
          double g = model == Model1 ? mu_g : exp(y[i]);
          cons_x[i-warmup] = log((gdp(y[i])+a[i]-g*a[i+1]*q(y[i],a[i+1])) * trend);
          tb_y_x[i-warmup] = (-a[i]+g*a[i+1]*q(y[i],a[i+1]))/gdp(y[i]);
          spread_x[i-warmup] = spread(y[i], a[i+1]);
        }
      else
        {
          cons_x[i-warmup] = log(gdp(y[i]) * trend);
          tb_y_x[i-warmup] = 0;
          spread_x[i-warmup] = 0;
        }

      if (model == Model1)
        trend *= mu_g;
      else
        trend *= exp(y[i]);

      debt_y_x[i-warmup] = -a[i]/gdp(y[i]);
      
      if (d[i+1] == Default)
        ndefaults++;
    }

  // HP filter
  double *gdp_x2 = new double[n-warmup-1];
  double *cons_x2 = new double[n-warmup-1];
  double *tb_y_x2 = new double[n-warmup-1];
  double *spread_x2 = new double[n-warmup-1];

  int v;
  v = hp_filter(gdp_x, n-warmup-1, 1600, gdp_x2, false);
  assert(v == 0);
  v = hp_filter(cons_x, n-warmup-1, 1600, cons_x2, false);
  assert(v == 0);
  v = hp_filter(tb_y_x, n-warmup-1, 1600, tb_y_x2, false);
  assert(v == 0);
  v = hp_filter(spread_x, n-warmup-1, 1600, spread_x2, false);
  assert(v == 0);

  r.sigma_y += gsl_stats_sd(gdp_x2, 1, n-warmup-1) * scale;
  r.sigma_c += gsl_stats_sd(cons_x2, 1, n-warmup-1) * scale;
  r.sigma_TB_y += gsl_stats_sd(tb_y_x2, 1, n-warmup-1) * scale;
  r.sigma_Rs += gsl_stats_sd(spread_x2, 1, n-warmup-1) * scale;
  r.rho_c_y += gsl_stats_correlation(cons_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_TB_y_y += gsl_stats_correlation(tb_y_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_Rs_y += gsl_stats_correlation(spread_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_Rs_TB_y += gsl_stats_correlation(spread_x2, 1, tb_y_x2, 1, n-warmup-1) * scale;
  r.E_Rs += gsl_stats_mean(spread_x, 1, n-warmup-1) * scale; // Not HP filtered
  r.defaults_10000 += ((double) ndefaults)/((double) (n-warmup-1))*10000.0 * scale;
  r.mean_debt_output += gsl_stats_mean(debt_y_x, 1, n-warmup-1) * scale; // Not HP filtered

  delete[] gdp_x;
  delete[] cons_x;
  delete[] tb_y_x;
  delete[] spread_x;
  delete[] debt_y_x;

  delete[] gdp_x2;
  delete[] cons_x2;
  delete[] tb_y_x2;
  delete[] spread_x2;
}

double
AguiarGopinath::difference(double y, void *params)
{
  difference_params *p = (difference_params *) params;
  const AguiarGopinath &AG = *p->AG;
  double a = p->a;
  double r = AG.VG(a, y) - AG.VB(y);
  return (r == -INFINITY ? almost_min_infinity : r);
}

double
AguiarGopinath::default_threshold(double y, double a1) const throw (LowerBound, UpperBound)
{
  double upper_y1, lower_y1;

  switch (model)
    {
    case Model1:
      upper_y1 = (1-rho_z)*mu_z + rho_z*y + width*sigma_z;
      lower_y1 = (1-rho_z)*mu_z + rho_z*y - width*sigma_z;
      break;
    case Model2:
      upper_y1 = (1-rho_g)*mu_y + rho_g*y + width*sigma_g;
      lower_y1 = (1-rho_g)*mu_y + rho_g*y - width*sigma_g;
      break;
    }

  if (VG(a1, lower_y1) >= VB(lower_y1))
    throw LowerBound();
  else if (VG(a1, upper_y1) <= VB(upper_y1))
    throw UpperBound();
  
  // Interior solution
  gsl_root_fsolver *s = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);

  gsl_function F;
  difference_params params = { a1, this };

  F.function = &AguiarGopinath::difference;
  F.params = &params;

  gsl_root_fsolver_set(s, &F, lower_y1, upper_y1);

  int status;
  do
    {
      status = gsl_root_fsolver_iterate(s);
      double x_lo = gsl_root_fsolver_x_lower(s);
      double x_hi = gsl_root_fsolver_x_upper(s);
      status = gsl_root_test_interval(x_lo, x_hi, root_tol, 0);
    }
  while (status == GSL_CONTINUE);

  double y1 = gsl_root_fsolver_root(s);

  gsl_root_fsolver_free(s);

  return y1;
}

AguiarGopinath::statistics_t
AguiarGopinath::model_statistics(size_t nsims, size_t simlength, size_t warmup) const
{
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r = gsl_rng_alloc(T);

  // Compute ergodic distribution
  std::cout << "Computing ergodic distribution..." << std::endl;
  size_t n = 10000;
  double *y = new double[n];
  double *a = new double[n];
  AguiarGopinath::state_t *d = new AguiarGopinath::state_t[n];

  y[0] = (model == AguiarGopinath::Model1 ? mu_z : mu_y);
  a[0] = 0;
  d[0] = AguiarGopinath::Repay;
  simulate(y, a, d, n, r);
  gsl_rng_free(r);

  double *y_ergodic = y+warmup;
  size_t n_ergodic = n-warmup;
  gsl_sort(y_ergodic, 1, n_ergodic);
  double *a_ergodic = a+warmup;
  gsl_sort(a_ergodic, 1, n_ergodic);

  // Draw 500 paths
  size_t n2 = simlength+1;

  AguiarGopinath::statistics_t st;

#pragma omp parallel shared(st)
  {
    double *y2 = new double[n2];
    double *a2 = new double[n2];
    AguiarGopinath::state_t *d2 = new AguiarGopinath::state_t[n2];
    gsl_rng *r2 = gsl_rng_alloc(T);
    gsl_rng_set(r2, omp_get_thread_num());

#pragma omp for schedule(dynamic)
    for (size_t i = 0; i < nsims; i++)
      {
        std::ostringstream msg;
        msg << "Simulation " << i+1 << "/" << nsims << "..." << std::endl;
        std::cout << msg.str();

        y2[0] = gsl_stats_quantile_from_sorted_data(y_ergodic, 1, n_ergodic,
                                                    gsl_rng_uniform(r2));
        a2[0] = gsl_stats_quantile_from_sorted_data(a_ergodic, 1, n_ergodic,
                                                    gsl_rng_uniform(r2));
        d2[0] = AguiarGopinath::Repay;
        simulate(y2, a2, d2, n2, r2);

#pragma omp critical
        simulation_statistics(y2, a2, d2, n2, 1000, st, 1.0/((double) nsims));

        if (i == 0)
          csv_output(y2, a2, d2, n2, std::string("yo.csv"));
      }

    delete[] y2;
    delete[] a2;
    delete[] d2;
    gsl_rng_free(r2);
  } // End #pragma parallel

  delete[] y;
  delete[] a;
  delete[] d;

  return st;
}

double
AguiarGopinath::VG_deriv(double a1, void *params)
{
  VG_deriv_params *p = (VG_deriv_params *) params;
  const AguiarGopinath &AG = *p->AG;
  double y = p->y;
  return AG.VG(a1, y);
}
