/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <ext/algorithm>

#include <gsl/gsl_deriv.h>
#include <gsl/gsl_sort.h>

#include "AguiarGopinath2EGM.hh"

AguiarGopinath2EGM::AguiarGopinath2EGM(model_t model_arg, double gam_arg, double beta_arg,
                                       double lambda_arg, double delta_arg, double convergence_tol_arg, size_t ngrid_a_arg, size_t ngrid_y_arg)
  : AguiarGopinath(model_arg, gam_arg, beta_arg, lambda_arg, delta_arg, convergence_tol_arg, ngrid_a_arg, ngrid_y_arg),
    VGint(ngrid_a, gsl_interp_cspline,
          ngrid_y, gsl_interp_cspline,
          false),
    BEVint(ngrid_a, gsl_interp_cspline,
           ngrid_y, gsl_interp_cspline),
    a1_dr_int(ngrid_a, gsl_interp_cspline,
              ngrid_y, gsl_interp_cspline,
              false)
{
  assert(ngrid_a == ngrid_a1);

  grid_a = new double[ngrid_a*ngrid_y];
  for (size_t iy = 0; iy < ngrid_y; iy++)
    __gnu_cxx::copy_n(grid_a1, ngrid_a, grid_a + iy * ngrid_a);

  // Initialize VB
  VBint.set_grid(grid_y);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    VBint.point(iy) = util(gdp(grid_y[iy])+min_a/3)*3;
  VBint.init();

  // Initialize VG
  VGint.set_grids(grid_a, grid_y);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      VGint.point(ia, iy) = util(gdp(grid_y[iy]) + grid_a[ia+iy*ngrid_a]/3)*3;
  VGint.init();

  // Initialize BEV
  init_BEV();

  // Initialize min_a1 et max_a1
  min_a1 = new double[ngrid_y];
  max_a1 = new double[ngrid_y];
  std::fill_n(min_a1, ngrid_y, min_a);
  std::fill_n(max_a1, ngrid_y, max_a);
}

AguiarGopinath2EGM::AguiarGopinath2EGM(const AguiarGopinath2EGM &arg)
  : AguiarGopinath(arg),
    VGint(arg.VGint),
    BEVint(arg.BEVint),
    a1_dr_int(arg.a1_dr_int)
{
  grid_a = new double[ngrid_a];
  __gnu_cxx::copy_n(arg.grid_a, ngrid_a, grid_a);
  min_a1 = new double[ngrid_y];
  __gnu_cxx::copy_n(arg.min_a1, ngrid_y, min_a1);
  max_a1 = new double[ngrid_y];
  __gnu_cxx::copy_n(arg.max_a1, ngrid_y, max_a1);
}

AguiarGopinath2EGM::~AguiarGopinath2EGM()
{
  delete[] grid_a;
  delete[] min_a1;
  delete[] max_a1;
}

int
AguiarGopinath2EGM::position_a1(double a1, double y, double *quad_points) const
{
  double z2 = NAN;
  double z1 = q(y, a1, &z2);

  double g = (model == Model1 ? mu_g : exp(y));

  double v2;
  if (!refine_pass)
    {
      gsl_function F;
      BEV_deriv_params params = { y, this };
      F.function = &AguiarGopinath2EGM::BEV_deriv;
      F.params = &params;

      double abserr;
      gsl_deriv_central(&F, a1, epsdiff, &v2, &abserr);
    }
  else
    {
      v2 = 0;
      for (size_t i = 0; i < gauss.npoints; i++)
        {
          double y1 = quad_points[i];
          if (!def(a1, y1))
            {
              double cons1 = cons(a1, y1, Repay);
              v2 += gauss.weight(i) * beta * pow(g*cons1, -gam)*g;
            }
        }
    }
  
  double cons = pow(v2/g/(z1+a1*z2), -1/gam);
  
  double a = cons + z1*a1*g - gdp(y);
  /*  
  std::cout << "Test: a1=" << a1 << ": " << (v2 <= 0 || v2/g/(z1+a1*z2) <= 0 || a < min_a ? -1 :
          (a > max_a ? 1 : 0)) << " (q=" << z1 << ",q'=" << z2 << ",BEV'=" << v2 << ",c=" << cons << ",a=" << a << ")" << std::endl;
  */
  return (v2 <= 0 || v2/g/(z1+a1*z2) <= 0 || a < min_a ? -1 :
          (a > max_a ? 1 : 0));
}


void
AguiarGopinath2EGM::compute()
{
  double *VB_points_new = new double[ngrid_y];
  double *VG_points_new = new double[ngrid_a*ngrid_y];
  double *BEV_points_old = new double[ngrid_a*ngrid_y];

  int it = 0;
  double err = INFINITY;

  refine_pass = false;
  
  while ((!refine_pass || err > convergence_tol) && it < MAX_IT)
    {
#pragma omp parallel shared(VB_points_new, VG_points_new)
      {
        double *quad_points = new double[gauss.npoints];
        
#pragma omp for schedule(dynamic)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          {
            double y = grid_y[iy];
            //          std::cout << iy+1 << " (y=" << y << ")" << std::endl;
            double g = (model == Model1 ? mu_g : exp(y));

            quadrature_points_y(y, quad_points);

            // Compute VB
            double EVB1 = 0;
            for (size_t i = 0; i < gauss.npoints; i++)
              EVB1 += gauss.weight(i)*VB(quad_points[i]);

            VB_points_new[iy] = util((1-delta)*gdp(y)) + lambda*BEV(0,y)
              + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

            // Compute VG
            assert(position_a1(max_a, y, quad_points) >= 0 && position_a1(min_a, y, quad_points) <= 0);

            /* double dichotomy_tol = (max_a1[iy]-min_a1[iy])/ngrid_a/2; */
            double dichotomy_tol = convergence_tol;

            // Find max.a1
            double max_a1_sup = max_a, max_a1_inf = min_a;
            while (max_a1_sup - max_a1_inf > dichotomy_tol)
              {
                double x = (max_a1_sup+max_a1_inf)/2;
                if (position_a1(x, y, quad_points) <= 0)
                  max_a1_inf = x;
                else
                  max_a1_sup = x;
              }
            max_a1[iy] = max_a1_inf;

            // Find min_a1
            double min_a1_sup = max_a, min_a1_inf = min_a;
            while (min_a1_sup - min_a1_inf > dichotomy_tol)
              {
                double x = (min_a1_sup+min_a1_inf)/2;
                if (position_a1(x, y, quad_points) >= 0)
                  min_a1_sup = x;
                else
                  min_a1_inf = x;
              }
            min_a1[iy] = min_a1_sup;

            assert(position_a1(max_a1[iy], y, quad_points) == 0
                   && position_a1(max_a1[iy], y, quad_points) == 0
                   && min_a1[iy] < max_a1[iy]);
            
            for (size_t ia = 0; ia < ngrid_a1; ia++)
              {
                double a1 = min_a1[iy]+gsl_pow_3(((double) ia)/((double) (ngrid_a1-1)))*(max_a1[iy]-min_a1[iy]);
              
                double z2 = NAN;
                double z1 = q(y, a1, &z2);

                double v1 = BEV(a1, y);

                double g = (model == Model1 ? mu_g : exp(y));

                double v2;

                if (!refine_pass)
                  {
                    gsl_function F;
                    BEV_deriv_params params = { y, this };
                    F.function = &AguiarGopinath2EGM::BEV_deriv;
                    F.params = &params;

                    double abserr;
                    gsl_deriv_central(&F, a1, epsdiff, &v2, &abserr);
                  }
                else
                  {
                    v2 = 0;
                    for (size_t i = 0; i < gauss.npoints; i++)
                      {
                        double y1 = quad_points[i];
                        if (!def(a1, y1))
                          {
                            double cons1 = cons(a1, y1, Repay);
                            v2 += gauss.weight(i) * beta * pow(g*cons1, -gam)*g;
                          }
                      }
                  }

                double cons = pow(v2/g/(z1+a1*z2), -1/gam);

                //!\\ Global writes
                grid_a[ia+iy*ngrid_a] = cons + z1*a1*g - gdp(y);
                VG_points_new[ia+iy*ngrid_a] = util(cons)+v1;
                /*
                  std::cout << std::fixed << std::setprecision(3)
                  << "a1=" << a1 << ", gdp=" << gdp(y) << ", q=" << z1
                  << ", q'=" << z2 << ", BEV'=" << v2 << ", c=" << cons
                  << ", a=" << grid_a[ia+iy*ngrid_a]
                  << ", VG=" << VG_points_new[ia+iy*ngrid_a] <<
                  std::endl;
                */
              }
          }
        
        delete[] quad_points;
      }
      

      err = 0;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        err = std::max(err, fabs(VB_points_new[iy]-VBint.point(iy)));
      VBint.init(VB_points_new);

      VGint.set_grids(grid_a, grid_y);
      VGint.init(VG_points_new);
      
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t ia = 0; ia < ngrid_a; ia++)
          BEV_points_old[ia+ngrid_a*iy] = BEV(grid_a1[ia],grid_y[iy]);

      init_BEV();
      
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t ia = 0; ia < ngrid_a; ia++)
          err = std::max(err, fabs(BEV_points_old[ia+iy*ngrid_a] - BEVint.point(ia,iy)));

      // Update min_a
      min_a = INFINITY;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        min_a = std::min(min_a, min_a1[iy]);
      min_a -= min_a_update_offset;
      
      for (size_t i = 0; i < ngrid_a1; i++)
        grid_a1[i] = min_a+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a);
      
      // Output
      std::cout << "Iteration " << ++it << ": error=" << std::scientific << err << ", min_a=" << min_a << std::endl;
      /*
      for (size_t ia = 0; ia < ngrid_a; ia++)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          std::cout << grid_a1[ia] << "," << grid_y[iy] << ": VG=" << VG(grid_a1[ia], grid_y[iy])
                    << ", BEV=" << BEV(grid_a1[ia], grid_y[iy]) << std::endl;
      */  

      if (!refine_pass && err < convergence_tol)
        {
          std::cout << "Moving to refine pass..." << std::endl;
          
          refine_pass = true;
          err = 1;
        }

      // Initialize a1_dr
      if (refine_pass)
        {
          a1_dr_int.set_grids(grid_a, grid_y);
          for (size_t iy = 0; iy < ngrid_y; iy++)
            for (size_t ia = 0; ia < ngrid_a1; ia++)
              a1_dr_int.point(ia, iy) = min_a1[iy]+gsl_pow_3(((double) ia)/((double) (ngrid_a1-1)))*(max_a1[iy]-min_a1[iy]);
          a1_dr_int.init();
        }
    }

  if (it == MAX_IT)
    std::cout << "Maximum number of iterations reached, aborting" << std::endl;
  else
    has_converged_flag = true;

  delete[] VB_points_new;
  delete[] VG_points_new;
  delete[] BEV_points_old;

}

void
AguiarGopinath2EGM::init_BEV()
{
  BEVint.set_grids(grid_a1, grid_y);

  double *quad_points = new double[gauss.npoints];

  for (size_t iy = 0; iy < ngrid_y; iy++)
    {
      double y = grid_y[iy];
      double g = (model == Model1 ? mu_g : exp(y));

      quadrature_points_y(y, quad_points);

      for (size_t ia = 0; ia < ngrid_a; ia++)
        {
          double a1 = grid_a1[ia];
          BEVint.point(ia, iy) = 0;
          for (size_t i = 0; i < gauss.npoints; i++)
            BEVint.point(ia, iy) += gauss.weight(i) * beta * pow(g, 1-gam) * V(a1, quad_points[i]);
        }
    }

  BEVint.init();
  
  delete[] quad_points;
}

double
AguiarGopinath2EGM::BEV_deriv(double a1, void *params)
{
  BEV_deriv_params *p = (BEV_deriv_params *) params;
  const AguiarGopinath2EGM &AG = *p->AG;
  double y = p->y;
  return AG.BEV(a1, y);
}
