/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <omp.h>

#include <gsl/gsl_min.h>

#include <iostream>

#include <ext/algorithm>

#include "AguiarGopinathVFI.hh"

AguiarGopinathVFI::AguiarGopinathVFI(model_t model_arg, double gam_arg, double beta_arg,
                                     double lambda_arg, double delta_arg, double convergence_tol_arg, size_t ngrid_a_arg, size_t ngrid_y_arg)
  : AguiarGopinath(model_arg, gam_arg, beta_arg, lambda_arg, delta_arg, convergence_tol_arg, ngrid_a_arg, ngrid_y_arg),
    VGint(ngrid_a, gsl_interp_cspline,
          ngrid_y, gsl_interp_cspline),
    a1_dr_int(ngrid_a, gsl_interp_cspline,
              ngrid_y, gsl_interp_cspline),
    maximization_tol(1e-6)
{
  grid_a = new double[ngrid_a];
  for (size_t i = 0; i < ngrid_a; i++)
    grid_a[i] = min_a+((double) i)/((double) (ngrid_a-1))*(max_a-min_a);
  
  // Initialize VB
  VBint.set_grid(grid_y);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    VBint.point(iy) = util((1-delta)*gdp(grid_y[iy]));
  VBint.init();

  // Initialize VG
  VGint.set_grids(grid_a, grid_y);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      VGint.point(ia, iy) = util(gdp(grid_y[iy]) + r * grid_a[ia]);
  VGint.init();

  // Initialize a1_dr
  a1_dr_int.set_grids(grid_a, grid_y);
}

AguiarGopinathVFI::AguiarGopinathVFI(const AguiarGopinathVFI &arg)
  : AguiarGopinath(arg),
    VGint(arg.VGint),
    a1_dr_int(arg.a1_dr_int),
    maximization_tol(arg.maximization_tol)
{
  grid_a = new double[ngrid_a];
  __gnu_cxx::copy_n(arg.grid_a, ngrid_a, grid_a);
}

AguiarGopinathVFI::~AguiarGopinathVFI()
{
  delete[] grid_a;
}

void
AguiarGopinathVFI::compute()
{
  double *VB_points_new = new double[ngrid_y];
  double *VG_points_new = new double[ngrid_a*ngrid_y];

  int it = 1;

  double err;

  do
    {
#pragma omp parallel shared(VB_points_new, VG_points_new)
      {
        double *quad_points = new double[gauss.npoints];
        double *EV1 = new double[ngrid_a1];
        gsl_min_fminimizer *minzer = gsl_min_fminimizer_alloc(gsl_min_fminimizer_brent);

#pragma omp for schedule(dynamic)
        for (size_t iy = 0; iy < ngrid_y; iy++)
          {
            double y = grid_y[iy];
            double g = (model == Model1 ? mu_g : exp(y));

            // Compute the quadrature points for y
            quadrature_points_y(y, quad_points);

            // Compute the quadrature points for tomorrow's VB, V and interest rate
            for (size_t ia1 = 0; ia1 < ngrid_a1; ia1++)
              {
                EV1[ia1] = 0.0;
                for (size_t i = 0; i < gauss.npoints; i++)
                  EV1[ia1] += gauss.weight(i) * V(grid_a1[ia1], quad_points[i]);
              }

            // Compute VB
            double EVB1 = 0.0, EVG1 = 0.0;
            for (size_t i = 0; i < gauss.npoints; i++)
              {
                EVB1 += gauss.weight(i)*VB(quad_points[i]);
                EVG1 += gauss.weight(i)*VG(0.0, quad_points[i]);
              }

            VB_points_new[iy] = util((1-delta)*gdp(y)) + lambda*beta*pow(g, 1-gam)*EVG1
              + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

            // Compute VG
            for (size_t ia = 0; ia < ngrid_a; ia++)
              {
                double a = grid_a[ia];
              
                // Global search
                double VGmax = -INFINITY;
                double a1_init = NAN;
                double min_a2 = INFINITY, max_a2 = -INFINITY;
                for (size_t ia1 = 0; ia1 < ngrid_a1; ia1++)
                  {
                    double a1 = grid_a1[ia1];
                    double c = gdp(y)+a-g*a1*q(y, a1);
                    if (c <= 0)
                      continue;
                    else
                      {
                        min_a2 = std::min(min_a2, a1);
                        max_a2 = std::max(max_a2, a1);
                      }
                    
                    double VGcur = util(c) + beta*pow(g, 1-gam)*EV1[ia1];
                    if (VGcur > VGmax)
                      {
                        VGmax = VGcur;
                        a1_init = a1;
                      }
                  }

                // Run the minimization
                if (VGmax != -INFINITY && (a1_init != min_a2 && a1_init != max_a2))
                  {
                    struct VG_params params = { a, y, g, quad_points, this };
                    gsl_function F;
                    F.function = &AguiarGopinathVFI::VG_objective;
                    F.params = &params;

                    gsl_min_fminimizer_set(minzer, &F, a1_init, min_a2, max_a2);

                    int status, it2 = 0;
                    do
                      {
                        status = gsl_min_fminimizer_iterate(minzer);
                        assert(status != GSL_EBADFUNC && status != GSL_FAILURE);

                        double a = gsl_min_fminimizer_x_lower(minzer);
                        double b = gsl_min_fminimizer_x_upper(minzer);
     
                        status = gsl_min_test_interval(a, b, maximization_tol, 0.0);
                      }
                    while (status == GSL_CONTINUE && ++it2 < MAX_IT);

                    VG_points_new[ia+iy*ngrid_a] = -gsl_min_fminimizer_f_minimum(minzer);
                    a1_dr_int.point(ia, iy) = gsl_min_fminimizer_x_minimum(minzer);
                  }
                else
                  {
                    VG_points_new[ia+iy*ngrid_a] = VGmax;
                    a1_dr_int.point(ia, iy) = (std::isnan(a1_init) ? -INFINITY : a1_init);
                  }
              }
          }
        gsl_min_fminimizer_free(minzer);
        delete[] quad_points;
        delete[] EV1;
      } // End #pragma parallel

      // Compute error
      err = 0;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        {
          err = std::max(err, fabs(VB_points_new[iy]-VBint.point(iy)));
          for (size_t ia = 0; ia < ngrid_a; ia++)
            {
              if (VG_points_new[ia+iy*ngrid_a] != -INFINITY || VGint.point(ia, iy) != -INFINITY)
                err = std::max(err, fabs(VG_points_new[ia+iy*ngrid_a]-VGint.point(ia, iy)));
              if ((VG_points_new[ia+iy*ngrid_a] == -INFINITY || VGint.point(ia, iy) == -INFINITY)
                  && VG_points_new[ia+iy*ngrid_a] != VGint.point(ia, iy))
                std::cout << "Point a=" << grid_a[ia] << ", y=" << grid_y[iy] << " switched to " << VG_points_new[ia+iy*ngrid_a] << std::endl;
            }
        }

      // Copy
      VBint.init(VB_points_new);
      VGint.init(VG_points_new);

      // Output
      std::cout << "Iteration " << it++ << ": error=" << err << std::endl;
    }
  while(err > convergence_tol && it < MAX_IT);

  if (it == MAX_IT)
    std::cout << "Maximum number of iterations reached, aborting" << std::endl;
  else
    has_converged_flag = true;

  delete[] VB_points_new;
  delete[] VG_points_new;

  a1_dr_int.init();
}

double
AguiarGopinathVFI::VG_objective(double a1, void *params)
{
  struct VG_params *p = (struct VG_params *) params;
  const AguiarGopinathVFI &AG = *p->AG;
  double y = p->y, a = p->a, g = p->g;

  double EV1 = 0;
  for (size_t i = 0; i < p->AG->gauss.npoints; i++)
    EV1 += AG.gauss.weight(i) * AG.V(a1, p->quad_points[i]);

  double v = -(AG.util(AG.gdp(y)+a-g*a1*AG.q(y, a1))
               + AG.beta*pow(g, 1-AG.gam)*EV1);

  if (std::isnan(v))
    return -almost_min_infinity;
  else
    return v;
}
