/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <boost/lambda/lambda.hpp>
#include <omp.h>

#include <gsl/gsl_sort.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_statistics.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_roots.h>
#include <gsl/gsl_deriv.h>

#include <iostream>
#include <sstream>
#include <iomanip>
#include <fstream>
#include <numeric>
#include <ext/algorithm>

#include "CohenVillemot.hh"
#include "hp_filter.hh"

CohenVillemot::CohenVillemot(double gam_arg, double beta_arg,
                             double lambda_arg, double delta_arg, double p_arg,
                             double q_arg, double mu_z_arg, double recovery_arg,
                             double convergence_tol_arg,
                             double min_a_arg, size_t ngrid_a1_arg)
  : gam(gam_arg), r(0.01), beta(beta_arg), lambda(lambda_arg),
    delta(delta_arg), mu_g(1.006),
    sigma_g(0.03), rho_g(0.17),
    mu_y(log(mu_g)-0.5*gsl_pow_2(sigma_g)/(1-gsl_pow_2(rho_g))),
    p(p_arg), q(q_arg), rho_z(0.8), mu_z(mu_z_arg),
    recovery(recovery_arg),
    min_a(min_a_arg), max_a(0.0),
    min_y(mu_y - 6*sigma_g/sqrt(1-gsl_pow_2(rho_g))),
    max_y(mu_y + 6*sigma_g/sqrt(1-gsl_pow_2(rho_g))),
    min_z(-2*mu_z), max_z(mu_z),
    ngrid_a(10), ngrid_y(10), ngrid_a1(ngrid_a1_arg), ngrid_z(10),
    width(6),
    VBCint(ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline),
    VBNint(ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline),
    gauss(sigma_g, 10, GaussQuadrature::Legendre),
    quad_npoints(2*gauss.npoints),
    convergence_tol(convergence_tol_arg),
    root_tol(1e-5)
{
  grid_y = new double[ngrid_y];
  for (size_t i = 0; i < ngrid_y; i++)
    grid_y[i] = min_y+((double) i)/((double) (ngrid_y-1))*(max_y-min_y);

  grid_z = new double[ngrid_z];
  for (size_t i = 0; i < ngrid_z; i++)
    grid_z[i] = min_z+((double) i)/((double) (ngrid_z-1))*(max_z-min_z);

  // All this needs to be in coordination with quadrature_points()
  quad_weights_normal = new double[quad_npoints];
  quad_weights_bonus_possible = new double[quad_npoints];
  quad_points_bonus_possible_normal = new bool[quad_npoints];
  quad_points_bonus_possible_bonus_possible = new bool[quad_npoints];
  for (size_t i = 0; i < gauss.npoints; i++)
      {
        quad_weights_normal[i] = gauss.weight(i) * (1-p);
        quad_weights_normal[i+gauss.npoints] = gauss.weight(i) * p;
        quad_points_bonus_possible_normal[i] = false;
        quad_points_bonus_possible_normal[i+gauss.npoints] = true;
        quad_weights_bonus_possible[i] = gauss.weight(i) * (1-q);
        quad_weights_bonus_possible[i+gauss.npoints] = gauss.weight(i) * q;
        quad_points_bonus_possible_bonus_possible[i] = true;
        quad_points_bonus_possible_bonus_possible[i+gauss.npoints] = false;
      }
}

CohenVillemot::CohenVillemot(const CohenVillemot &arg)
  : gam(arg.gam), r(arg.r), beta(arg.beta), lambda(arg.lambda),
    delta(arg.delta), mu_g(arg.mu_g), sigma_g(arg.sigma_g), rho_g(arg.rho_g), mu_y(arg.mu_y),
    p(arg.p), q(arg.q), rho_z(arg.rho_z), mu_z(arg.mu_z), recovery(arg.recovery),
    min_a(arg.min_a), max_a(arg.max_a), min_y(arg.min_y), max_y(arg.max_y),
    min_z(arg.min_z), max_z(arg.max_z),
    ngrid_a(arg.ngrid_a), ngrid_y(arg.ngrid_y), ngrid_a1(arg.ngrid_a1),
    ngrid_z(arg.ngrid_z),
    width(arg.width),
    VBCint(arg.VBCint),
    VBNint(arg.VBCint),
    gauss(arg.gauss),
    quad_npoints(arg.quad_npoints), convergence_tol(arg.convergence_tol),
    root_tol(arg.root_tol)
{
  grid_y = new double[ngrid_y];
  __gnu_cxx::copy_n(arg.grid_y, ngrid_y, grid_y);
  grid_z = new double[ngrid_z];
  __gnu_cxx::copy_n(arg.grid_z, ngrid_z, grid_z);  

  quad_weights_normal = new double[quad_npoints];
  __gnu_cxx::copy_n(arg.quad_weights_normal, quad_npoints, quad_weights_normal);
  quad_weights_bonus_possible = new double[quad_npoints];
  __gnu_cxx::copy_n(arg.quad_weights_bonus_possible, quad_npoints, quad_weights_bonus_possible);
  quad_points_bonus_possible_normal = new bool[quad_npoints];
  __gnu_cxx::copy_n(arg.quad_points_bonus_possible_normal, quad_npoints, quad_points_bonus_possible_normal);
  quad_points_bonus_possible_bonus_possible = new bool[quad_npoints];
  __gnu_cxx::copy_n(arg.quad_points_bonus_possible_bonus_possible, quad_npoints, quad_points_bonus_possible_bonus_possible);
}

CohenVillemot::~CohenVillemot()
{
  delete[] grid_y;
  delete[] grid_z;

  delete[] quad_weights_normal;
  delete[] quad_weights_bonus_possible;
  delete[] quad_points_bonus_possible_normal;
  delete[] quad_points_bonus_possible_bonus_possible;
}

double
CohenVillemot::VBC(double y, double z) const
{
  return VBCint.get(y, z);
}

double
CohenVillemot::VBN(double y, double z) const
{
  return VBNint.get(y, z);
}

double
CohenVillemot::price(double y, double z, bool bonus_possible, double a1, double *deriv_a1) const
{
  assert(!std::isnan(y) && !std::isnan(z) && !std::isnan(a1));

  if (a1 >= 0)
    {
      if (deriv_a1 != NULL)
        *deriv_a1 = 0;
      
      return 1.0/(1.0+r);
    }

  double npi1, npi2, npi1_deriv, npi2_deriv;
  if (bonus_possible)
    {
      npi1 = repayment_probability(y, rho_z*z, a1, true, &npi1_deriv);
      npi2 = repayment_probability(y, rho_z*z+mu_z, a1, false, &npi2_deriv);

      if (deriv_a1 != NULL)
        *deriv_a1 = ((1-q)*(npi1_deriv*(1-recovery/a1)-(1-npi1)*recovery/gsl_pow_2(a1))
                     + q*(npi2_deriv*(1-recovery/a1)-(1-npi2)*recovery/gsl_pow_2(a1)))/(1.0+r);
      
      // NB: both recovery and a1 are negative
      return ((1-q)*(npi1+(1-npi1)*recovery/a1)
              + q*(npi2+(1-npi2)*recovery/a1))/(1.0+r);
    }
  else
    {
      npi1 = repayment_probability(y, rho_z*z, a1, false, &npi1_deriv);
      npi2 = repayment_probability(y, rho_z*z-mu_z, a1, true, &npi2_deriv);

      if (deriv_a1 != NULL)
        *deriv_a1 = ((1-p)*(npi1_deriv*(1-recovery/a1)-(1-npi1)*recovery/gsl_pow_2(a1))
                     + p*(npi2_deriv*(1-recovery/a1)-(1-npi2)*recovery/gsl_pow_2(a1)))/(1.0+r);
        
      // NB: both recovery and a1 are negative
      return ((1-p)*(npi1+(1-npi1)*recovery/a1)
              + p*(npi2+(1-npi2)*recovery/a1))/(1.0+r);
    }
}

double
CohenVillemot::crisis_prob(double y, double z, bool bonus_possible, double a1) const
{
  assert(!std::isnan(y) && !std::isnan(z) && !std::isnan(a1));

  if (a1 >= 0)
    return 0;

  double npi1, npi2, prob;
  if (bonus_possible)
    {
      npi1 = repayment_probability(y, rho_z*z, a1, true);
      npi2 = repayment_probability(y, rho_z*z+mu_z, a1, false);
      prob = (1-q)*npi1 + q*npi2;
    }
  else
    {
      npi1 = repayment_probability(y, rho_z*z, a1, false);
      npi2 = repayment_probability(y, rho_z*z-mu_z, a1, true);
      prob = (1-p)*npi1 + p*npi2; 
    }
  return gsl_pow_4(1/prob)-1;
}

void
CohenVillemot::quadrature_points_B(double y, double z, double *quad_points_y1) const
{
  std::transform(gauss.get_points(), gauss.get_points() + gauss.npoints, quad_points_y1,
                 boost::lambda::_1+(1-rho_g)*mu_y + rho_g*y);
}

void
CohenVillemot::quadrature_points_G(double y, double z, bool bonus_possible,
                                   double *quad_points_y1, double *quad_points_z1) const
{
  quadrature_points_B(y, z, quad_points_y1);
  
  __gnu_cxx::copy_n(quad_points_y1, gauss.npoints, quad_points_y1 + gauss.npoints);

  std::fill_n(quad_points_z1, gauss.npoints, rho_z*z);
  if (bonus_possible)
    std::fill_n(quad_points_z1 + gauss.npoints, gauss.npoints, rho_z*z+mu_z);
  else
    std::fill_n(quad_points_z1 + gauss.npoints, gauss.npoints, rho_z*z-mu_z);

}

double
CohenVillemot::spread(double y, double z, bool bonus_possible, double a1) const
{
  assert(!std::isnan(y) && !std::isnan(a1) && !std::isnan(z));

  return gsl_pow_4(1/price(y, z, bonus_possible, a1)/(1.0+r)) - 1.0;
}

double
CohenVillemot::cons(double a, double y, double z, bool bonus_possible, state_t decision) const
{
  assert(!std::isnan(y) && !std::isnan(a) && !std::isnan(z));

  double g = exp(y)+z;
  double a1;
  switch (decision)
    {
    case Default:
    case WaitRedemption:
      return (1-delta)*gdp(y, z);
    case Redemption:
    case Repay:
      a1 = a1_dr(a, y, z, bonus_possible);
      return gdp(y, z)+a-g*a1*price(y, z, bonus_possible, a1);
    }
}

void
CohenVillemot::simulate(double *a, double *y, double *z, state_t *d, bool *bonus_possible, size_t n, gsl_rng *rng) const
{
  for (size_t i = 0; i < n-1; i++)
    {
      y[i+1] = (1-rho_g)*mu_y+rho_g*y[i]+gsl_ran_gaussian(rng, sigma_g);

      z[i+1] = rho_z*z[i];
      
      if (d[i] == Default || d[i] == WaitRedemption)
        {
          if (gsl_rng_uniform(rng) < lambda)
            {
              d[i+1] = Redemption;
              a[i+1] = a1_dr(recovery, y[i], z[i], false);
              if (gsl_rng_uniform(rng) < p)
                {
                  z[i+1] -= mu_z;
                  bonus_possible[i+1] = true;
                }
              else
                bonus_possible[i+1] = false;
            }
          else
            {
              d[i+1] = WaitRedemption;
              a[i+1] = recovery;
              bonus_possible[i+1] = false;
            }
        }
      else
        {
          if (def(a[i], y[i], z[i], bonus_possible[i]))
            {
              d[i+1] = Default;
              a[i+1] = recovery;
              if (!bonus_possible[i])
                z[i+1] -= mu_z;
              bonus_possible[i+1] = false;
            }
          else
            {
              d[i+1] = Repay;
              a[i+1] = a1_dr(a[i], y[i], z[i], bonus_possible[i]);
              if (bonus_possible[i])
                if (gsl_rng_uniform(rng) < q)
                  {
                    z[i+1] += mu_z;
                    bonus_possible[i+1] = false;
                  }
                else
                  bonus_possible[i+1] = true;
              else
                if (gsl_rng_uniform(rng) < p)
                  {
                    z[i+1] -= mu_z;
                    bonus_possible[i+1] = true;
                  }
                else
                  bonus_possible[i+1] = false;
            }
        }
    }
}

void
CohenVillemot::print_simulation(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n) const
{
  for (size_t i = 0; i < n-1; i++)
    {
      std::cout << "State: y=" << gdp(y[i], z[i]) << (bonus_possible[i] ? "bonus possible" : "no bonus possible")
                << ", a=" << a[i] << ", credit history=";
      switch (d[i])
        {
        case Default:
        case WaitRedemption:
          std::cout << "bad";
          break;
        case Repay:
        case Redemption:
          std::cout << "good";
          break;
        }
      std::cout << "; Decision: c=" << cons(a[i], y[i], z[i], bonus_possible[i], d[i+1]) << ", ";
      switch (d[i+1])
        {
        case Default:
          std::cout << "default, VB=" << (bonus_possible[i] ? VBC(y[i], z[i]) : VBN(y[i], z[i]));
          break;
        case WaitRedemption:
          std::cout << "wait redemption, VB=" << VBC(y[i], z[i]);
          break;
        case Repay:
          std::cout << "repay, a1=" << a[i+1] << ", rs=" << spread(y[i], z[i], bonus_possible[i], a[i+1]) << ", V=" << (bonus_possible[i] ? VGC(a[i], y[i], z[i]) : VGN(a[i], y[i], z[i]));
          break;
        case Redemption:
          std::cout << "redemption, a1=" << a[i+1] << ", rs=" << spread(y[i], z[i], bonus_possible[i], a[i+1]) << ", V=" << VGN(recovery, y[i], z[i]);
          break;
        }
      std::cout << std::endl;
    }
}

void
CohenVillemot::csv_output(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n, const std::string &filename) const
{
  std::ofstream output(filename.c_str());

  output << std::scientific << std::setprecision(10);

  output << "y,z,gdp,a,c,d,a1,rs,V,bonus_possible" << std::endl;
  for (size_t i = 0; i < n-1; i++)
    {
      output << y[i] << "," << z[i] << "," << gdp(y[i], z[i]) << "," << a[i] << "," << cons(a[i], y[i], z[i], bonus_possible[i], d[i+1]) << ",";

      switch (d[i+1])
        {
        case Default:
          output << "default,nan," << recovery << "," << (bonus_possible[i] ? VBC(y[i], z[i]) : VBN(y[i], z[i]));
          break;
        case WaitRedemption:
          output << "wait_redemption,nan," << recovery << "," << VBC(y[i], z[i]);
          break;
        case Repay:
          output << "repay," << a[i+1] << "," << spread(y[i], z[i], bonus_possible[i], a[i+1]) << "," << (bonus_possible[i] ? VGC(a[i], y[i], z[i]) : VGN(a[i], y[i], z[i]));
          break;
        case Redemption:
          output << "redemption," << a[i+1] << "," << spread(y[i], z[i], bonus_possible[i], a[i+1]) << "," << VGN(recovery, y[i], z[i]);
          break;
        }
      output << "," << bonus_possible[i] << std::endl;
    }

  output.close();
}

void
CohenVillemot::simulation_statistics(const double *a, const double *y, const double *z, const state_t *d, const bool *bonus_possible, size_t n, size_t warmup, statistics_t &r, double scale) const
{
  assert(n > warmup+1);

  double *gdp_x = new double[n-warmup-1];
  double *cons_x = new double[n-warmup-1];
  double *tb_y_x = new double[n-warmup-1];
  double *spread_x = new double[n-warmup-1];
  double *debt_y_x = new double[n-warmup-1];
  double *welfare_x = new double[n-warmup-1];
  size_t ndefaults = 0;

  double trend = mu_g;

  for (size_t i = warmup; i < n-1; i++)
    {
      gdp_x[i-warmup] = log(gdp(y[i], z[i]) * trend);

      if (d[i+1] == Repay || d[i+1] == Redemption)
        {
          double g = exp(y[i])+z[i];
          cons_x[i-warmup] = log((gdp(y[i], z[i])+a[i]-g*a[i+1]*price(y[i],z[i],bonus_possible[i],a[i+1])) * trend);
          tb_y_x[i-warmup] = (-a[i]+g*a[i+1]*price(y[i],z[i],bonus_possible[i],a[i+1]))/gdp(y[i],z[i]);
          spread_x[i-warmup] = spread(y[i], z[i], bonus_possible[i], a[i+1]);
        }
      else
        {
          cons_x[i-warmup] = log(gdp(y[i], z[i]) * trend);
          tb_y_x[i-warmup] = 0;
          spread_x[i-warmup] = 0;
        }

      trend *= exp(y[i])+z[i];

      debt_y_x[i-warmup] = -a[i]/gdp(y[i], z[i]);
      welfare_x[i-warmup] = V(a[i], y[i], z[i], bonus_possible[i]);
      
      if (d[i+1] == Default)
        ndefaults++;
    }

  // HP filter
  double *gdp_x2 = new double[n-warmup-1];
  double *cons_x2 = new double[n-warmup-1];
  double *tb_y_x2 = new double[n-warmup-1];
  double *spread_x2 = new double[n-warmup-1];

  int v;
  v = hp_filter(gdp_x, n-warmup-1, 1600, gdp_x2, false);
  assert(v == 0);
  v = hp_filter(cons_x, n-warmup-1, 1600, cons_x2, false);
  assert(v == 0);
  v = hp_filter(tb_y_x, n-warmup-1, 1600, tb_y_x2, false);
  assert(v == 0);
  v = hp_filter(spread_x, n-warmup-1, 1600, spread_x2, false);
  assert(v == 0);

  r.sigma_y += gsl_stats_sd(gdp_x2, 1, n-warmup-1) * scale;
  r.sigma_c += gsl_stats_sd(cons_x2, 1, n-warmup-1) * scale;
  r.sigma_TB_y += gsl_stats_sd(tb_y_x2, 1, n-warmup-1) * scale;
  r.sigma_Rs += gsl_stats_sd(spread_x2, 1, n-warmup-1) * scale;
  r.rho_c_y += gsl_stats_correlation(cons_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_TB_y_y += gsl_stats_correlation(tb_y_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_Rs_y += gsl_stats_correlation(spread_x2, 1, gdp_x2, 1, n-warmup-1) * scale;
  r.rho_Rs_TB_y += gsl_stats_correlation(spread_x2, 1, tb_y_x2, 1, n-warmup-1) * scale;
  r.E_Rs += gsl_stats_mean(spread_x, 1, n-warmup-1) * scale; // Not HP filtered
  r.defaults_10000 += ((double) ndefaults)/((double) (n-warmup-1))*10000.0 * scale;
  r.mean_debt_output += gsl_stats_mean(debt_y_x, 1, n-warmup-1) * scale; // Not HP filtered
  r.mean_welfare += gsl_stats_mean(welfare_x, 1, n-warmup-1) * scale; // Not HP filtered

  delete[] gdp_x;
  delete[] cons_x;
  delete[] tb_y_x;
  delete[] spread_x;
  delete[] debt_y_x;
  delete[] welfare_x;

  delete[] gdp_x2;
  delete[] cons_x2;
  delete[] tb_y_x2;
  delete[] spread_x2;
}

double
CohenVillemot::difference(double y, void *params)
{
  difference_params *p = (difference_params *) params;
  const CohenVillemot &CV = *p->CV;
  double a = p->a, z = p->z;
  if (p->bonus_possible)
    return CV.VGC(a, y, z) - CV.VBC(y, z);
  else
    return CV.VGN(a, y, z) - CV.VBN(y, z);
}

double
CohenVillemot::repayment_probability(double y, double z1, double a1, bool bonus_possible1, double *deriv_a1) const
{
  double cond_mean = (1-rho_g)*mu_y + rho_g*y;

  double upper_y1 = cond_mean + width*sigma_g,
    lower_y1 = cond_mean - width*sigma_g;

  while ((VGN(a1, lower_y1, z1) == -INFINITY || VGC(a1, lower_y1, z1) == -INFINITY)
         && lower_y1 < upper_y1)
    lower_y1 += 0.005;

  difference_params params = { a1, z1, bonus_possible1, this };

  if (difference(lower_y1, &params) >= 0)
    {
      if (deriv_a1 != NULL)
        *deriv_a1 = 0;
      
      return 1.0;
    }
  else if (difference(upper_y1, &params) <= 0)
    {
      if (deriv_a1 != NULL)
        *deriv_a1 = 0;
      
      return 0.0;
    }

  assert(std::isfinite(VGN(a1, upper_y1, z1)));
  assert(std::isfinite(VGN(a1, lower_y1, z1)));
  assert(std::isfinite(VGC(a1, upper_y1, z1)));
  assert(std::isfinite(VGC(a1, lower_y1, z1)));
  
  // Interior solution

  gsl_root_fsolver *s = gsl_root_fsolver_alloc(gsl_root_fsolver_brent);

  gsl_function F;

  F.function = &CohenVillemot::difference;
  F.params = &params;

  gsl_root_fsolver_set(s, &F, lower_y1, upper_y1);

  int status;
  do
    {
      status = gsl_root_fsolver_iterate(s);
      double x_lo = gsl_root_fsolver_x_lower(s);
      double x_hi = gsl_root_fsolver_x_upper(s);
      status = gsl_root_test_interval(x_lo, x_hi, root_tol, 0);
    }
  while (status == GSL_CONTINUE);

  double y1 = gsl_root_fsolver_root(s);

  gsl_root_fsolver_free(s);

  if (deriv_a1 != NULL)
    {
      gsl_function F;
      double abserr, VG_deriv_a;
      VG_deriv_params par = { y1, z1, bonus_possible1, this };
      F.function = &CohenVillemot::VG_deriv;
      F.params = &par;
      gsl_deriv_central(&F, a1, epsdiff, &VG_deriv_a, &abserr);

      F.function = &CohenVillemot::difference;
      F.params = &params;
      double difference_diff;
      gsl_deriv_central(&F, y1, epsdiff, &difference_diff, &abserr);

      if (std::isfinite(VG_deriv_a) && std::isfinite(difference_diff))
        *deriv_a1 = gsl_ran_gaussian_pdf(y1 - cond_mean, sigma_g)
          *VG_deriv_a/difference_diff;
      else
        *deriv_a1 = NAN;
    }

  return gsl_cdf_gaussian_Q(y1 - cond_mean, sigma_g);
}

CohenVillemot::statistics_t
CohenVillemot::model_statistics(size_t nsims, size_t simlength, size_t warmup) const
{
  const gsl_rng_type *T = gsl_rng_default;
  gsl_rng *r = gsl_rng_alloc(T);

  // Compute ergodic distribution
  std::cout << "Computing ergodic distribution..." << std::endl;
  size_t n = 10000;
  double *y = new double[n];
  double *z = new double[n];
  double *a = new double[n];
  CohenVillemot::state_t *d = new CohenVillemot::state_t[n];
  bool *bonus_possible = new bool[n];
  
  y[0] = mu_y;
  z[0] = 0;
  a[0] = 0;
  d[0] = CohenVillemot::Repay;
  bonus_possible[0] = false;
  simulate(a, y, z, d, bonus_possible, n, r);
  gsl_rng_free(r);

  double *y_ergodic = y+warmup;
  double *z_ergodic = z+warmup;
  size_t n_ergodic = n-warmup;
  gsl_sort(y_ergodic, 1, n_ergodic);
  gsl_sort(z_ergodic, 1, n_ergodic);
  double *a_ergodic = a+warmup;
  gsl_sort(a_ergodic, 1, n_ergodic);

  // Draw 500 paths
  size_t n2 = simlength+1;

  CohenVillemot::statistics_t st;

#pragma omp parallel shared(st)
  {
    double *y2 = new double[n2];
    double *z2 = new double[n2];
    double *a2 = new double[n2];
    CohenVillemot::state_t *d2 = new CohenVillemot::state_t[n2];
    bool *bonus_possible2 = new bool[n2];
    gsl_rng *r2 = gsl_rng_alloc(T);
    gsl_rng_set(r2, omp_get_thread_num());

#pragma omp for schedule(dynamic)
    for (size_t i = 0; i < nsims; i++)
      {
        std::ostringstream msg;
        msg << "Simulation " << i+1 << "/" << nsims << "..." << std::endl;
        std::cout << msg.str();

        y2[0] = gsl_stats_quantile_from_sorted_data(y_ergodic, 1, n_ergodic,
                                                    gsl_rng_uniform(r2));
        z2[0] = gsl_stats_quantile_from_sorted_data(z_ergodic, 1, n_ergodic,
                                                    gsl_rng_uniform(r2));
        a2[0] = gsl_stats_quantile_from_sorted_data(a_ergodic, 1, n_ergodic,
                                                    gsl_rng_uniform(r2));
        d2[0] = CohenVillemot::Repay;
        bonus_possible2[0] = false;
        simulate(a2, y2, z2, d2, bonus_possible2, n2, r2);

#pragma omp critical
        simulation_statistics(a2, y2, z2, d2, bonus_possible2, n2, 1000, st, 1.0/((double) nsims));

        if (i == 0)
          csv_output(a2, y2, z2, d2, bonus_possible2, n2, std::string("yo.csv"));
      }

    delete[] y2;
    delete[] z2;
    delete[] a2;
    delete[] d2;
    delete[] bonus_possible2;
    gsl_rng_free(r2);
  } // End #pragma parallel

  delete[] y;
  delete[] z;
  delete[] a;
  delete[] d;
  delete[] bonus_possible;

  return st;
}

double
CohenVillemot::VG_deriv(double a, void *params)
{
  VG_deriv_params *p = (VG_deriv_params *) params;
  const CohenVillemot &CV = *p->CV;
  double y = p->y, z = p->z;
  if (p->bonus_possible)
    return CV.VGC(a, y, z);
  else
    return CV.VGN(a, y, z);
}

