/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <ext/algorithm>
#include <gsl/gsl_min.h>

#include <iostream>
#include <cassert>

#include "CohenVillemotVFI.hh"

CohenVillemotVFI::CohenVillemotVFI(double gam_arg, double beta_arg,
                                   double lambda_arg, double delta_arg, double p_arg,
                                   double q_arg, double mu_z_arg, double recovery_arg,
                                   double convergence_tol_arg, double min_a_arg)
  : CohenVillemot(gam_arg, beta_arg, lambda_arg, delta_arg, p_arg, q_arg, mu_z_arg,
                  recovery_arg, convergence_tol_arg, min_a_arg, 20),
    VGNint(ngrid_a, gsl_interp_cspline,
           ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline),
    VGCint(ngrid_a, gsl_interp_cspline,
           ngrid_y, gsl_interp_cspline,
           ngrid_z, gsl_interp_cspline),
    a1_dr_normal_int(ngrid_a, gsl_interp_cspline,
                     ngrid_y, gsl_interp_cspline,
                     ngrid_z, gsl_interp_cspline),
    a1_dr_bonus_possible_int(ngrid_a, gsl_interp_cspline,
                             ngrid_y, gsl_interp_cspline,
                             ngrid_z, gsl_interp_cspline),
    maximization_tol(1e-6)
{
  grid_a = new double[ngrid_a];
  for (size_t i = 0; i < ngrid_a; i++)
    grid_a[i] = min_a+((double) i)/((double) (ngrid_a-1))*(max_a-min_a);
  
  grid_a1 = new double[ngrid_a1];
  for (size_t i = 0; i < ngrid_a1; i++)
    grid_a1[i] = min_a+((double) i)/((double) (ngrid_a1-1))*(max_a-min_a);

  // Initialize VBN
  VBNint.set_grids(grid_y, grid_z);
  for (size_t iz = 0; iz < ngrid_z; iz++)
    for (size_t iy = 0; iy < ngrid_y; iy++)
      VBNint.point(iy, iz) = util((1-delta)*gdp(grid_y[iy], grid_z[iz]));
  VBNint.init();

  // Initialize VBC
  VBCint.set_grids(grid_y, grid_z);
  for (size_t iz = 0; iz < ngrid_z; iz++)
    for (size_t iy = 0; iy < ngrid_y; iy++)
      VBCint.point(iy, iz) = util((1-delta)*gdp(grid_y[iy], grid_z[iz]));
  VBCint.init();

  // Initialize VGN
  VGNint.set_grids(grid_a, grid_y, grid_z);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      for (size_t iz = 0; iz < ngrid_z; iz++)
        VGNint.point(ia, iy, iz) = util(gdp(grid_y[iy], grid_z[iz]) + r * grid_a[ia]);
  VGNint.init();
  
  // Initialize VGC
  VGCint.set_grids(grid_a, grid_y, grid_z);
  for (size_t iy = 0; iy < ngrid_y; iy++)
    for (size_t ia = 0; ia < ngrid_a; ia++)
      for (size_t iz = 0; iz < ngrid_z; iz++)
        VGCint.point(ia, iy, iz) = util(gdp(grid_y[iy], grid_z[iz]) + r * grid_a[ia]);
  VGCint.init();

  // Initialize a1_dr
  a1_dr_normal_int.set_grids(grid_a, grid_y, grid_z);
  a1_dr_bonus_possible_int.set_grids(grid_a, grid_y, grid_z);
}

CohenVillemotVFI::CohenVillemotVFI(const CohenVillemotVFI &arg)
  : CohenVillemot(arg),
    VGNint(arg.VGNint),
    VGCint(arg.VGCint),
    a1_dr_normal_int(arg.a1_dr_normal_int),
    a1_dr_bonus_possible_int(arg.a1_dr_bonus_possible_int),
    maximization_tol(arg.maximization_tol)
{
  grid_a = new double[ngrid_a];
  __gnu_cxx::copy_n(arg.grid_a, ngrid_a, grid_a);
  grid_a1 = new double[ngrid_a1];
  __gnu_cxx::copy_n(arg.grid_a1, ngrid_a1, grid_a1);
}

CohenVillemotVFI::~CohenVillemotVFI()
{
  delete[] grid_a;
  delete[] grid_a1;
}

void
CohenVillemotVFI::compute() throw (ConvergenceError)
{
  double *VBN_points_new = new double[ngrid_y*ngrid_z];
  double *VBC_points_new = new double[ngrid_y*ngrid_z];
  double *VGN_points_new = new double[ngrid_a*ngrid_y*ngrid_z];
  double *VGC_points_new = new double[ngrid_a*ngrid_y*ngrid_z];

  int it = 1;

  double err;

  do
    {
#pragma omp parallel shared(VBN_points_new, VBC_points_new, VGN_points_new, VGC_points_new)
      {
        double
          *quad_points_y1_normal = new double[quad_npoints],
          *quad_points_z1_normal = new double[quad_npoints],
          *quad_points_y1_bonus_possible = new double[quad_npoints],
          *quad_points_z1_bonus_possible = new double[quad_npoints],
          *quad_points_y1_VB = new double[gauss.npoints];
        double *EV1_normal = new double[ngrid_a1], *EV1_bp = new double[ngrid_a1];
        gsl_min_fminimizer *minzer = gsl_min_fminimizer_alloc(gsl_min_fminimizer_brent);

        for (size_t iy = 0; iy < ngrid_y; iy++)
          for (size_t iz = 0; iz < ngrid_z; iz++)
            {
              double y = grid_y[iy];
              double z = grid_z[iz];
              double g = exp(y)+z;

              // Compute the quadrature points for y
              quadrature_points_B(y, z, quad_points_y1_VB);
              quadrature_points_G(y, z, false, quad_points_y1_normal, quad_points_z1_normal);
              quadrature_points_G(y, z, true, quad_points_y1_bonus_possible, quad_points_z1_bonus_possible);

              // Compute the quadrature points for tomorrow's VB, V and interest rate
              for (size_t ia1 = 0; ia1 < ngrid_a1; ia1++)
                {
                  EV1_normal[ia1] = 0.0;
                  EV1_bp[ia1] = 0.0;
                  for (size_t i = 0; i < quad_npoints; i++)
                    {
                      EV1_normal[ia1] += quad_weights_normal[i] * V(grid_a1[ia1], quad_points_y1_normal[i], quad_points_z1_normal[i], quad_points_bonus_possible_normal[i]);
                      EV1_bp[ia1] += quad_weights_bonus_possible[i] * V(grid_a1[ia1], quad_points_y1_bonus_possible[i], quad_points_z1_bonus_possible[i], quad_points_bonus_possible_bonus_possible[i]);
                    }
                }

              // Compute VBN and VBC
              double EVB1 = 0.0, EVG1 = 0.0;
              for (size_t i = 0; i < gauss.npoints; i++)
                {
                  EVB1 += gauss.weight(i)*VBC(quad_points_y1_VB[i], rho_z*z);
                  EVG1 += gauss.weight(i)*VGN(recovery, quad_points_y1_VB[i], rho_z*z);
                }

              VBC_points_new[iy+iz*ngrid_y] = util((1-delta)*gdp(y, z)) + lambda*beta*pow(g, 1-gam)*EVG1
                + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

              EVB1 = 0.0;
              EVG1 = 0.0;
              for (size_t i = 0; i < gauss.npoints; i++)
                {
                  EVB1 += gauss.weight(i)*VBC(quad_points_y1_VB[i], rho_z*z-mu_z);
                  EVG1 += gauss.weight(i)*VGN(recovery, quad_points_y1_VB[i], rho_z*z-mu_z);
                }

              VBN_points_new[iy+iz*ngrid_y] = util((1-delta)*gdp(y, z)) + lambda*beta*pow(g, 1-gam)*EVG1
                + (1-lambda)*beta*pow(g, 1-gam)*EVB1;

              // Compute VGN and VGC
#pragma omp for schedule(dynamic)
              for (size_t ia = 0; ia < ngrid_a; ia++)
                {
                  double a = grid_a[ia];
              
                  // Global search
                  double VGNmax = -INFINITY;
                  double a1_init = NAN;
                  double min_a2 = INFINITY, max_a2 = -INFINITY;

                  for (size_t ia1 = 0; ia1 < ngrid_a1; ia1++)
                    {
                      double a1 = grid_a1[ia1];
                      double c = gdp(y, z)+a-g*a1*price(y, z, false, a1);
                      
                      if (c <= 0)
                        continue;
                      else
                        {
                          min_a2 = std::min(min_a2, a1);
                          max_a2 = std::max(max_a2, a1);
                        }

                      double VGNcur = util(c) + beta*pow(g, 1-gam)*EV1_normal[ia1];
                      if (VGNcur > VGNmax)
                        {
                          VGNmax = VGNcur;
                          a1_init = a1;
                        }
                    }

                  // Run the minimization
                  if (VGNmax != -INFINITY && (a1_init != min_a2 && a1_init != max_a2))
                    {
                      struct VG_params params = { a, y, z, quad_points_y1_normal, quad_points_z1_normal, this };
                      gsl_function F;
                      F.function = &CohenVillemotVFI::VGN_objective;
                      F.params = &params;

                      gsl_min_fminimizer_set(minzer, &F, a1_init, min_a2, max_a2);

                      int status, it2 = 0;
                      do
                        {
                          status = gsl_min_fminimizer_iterate(minzer);
                          assert(status != GSL_EBADFUNC && status != GSL_FAILURE);

                          double x1 = gsl_min_fminimizer_x_lower(minzer);
                          double x2 = gsl_min_fminimizer_x_upper(minzer);
     
                          status = gsl_min_test_interval(x1, x2, maximization_tol, 0.0);
                        }
                      while (status == GSL_CONTINUE && ++it2 < MAX_IT);
                      
                      VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] = -gsl_min_fminimizer_f_minimum(minzer);
                      a1_dr_normal_int.point(ia, iy, iz) = gsl_min_fminimizer_x_minimum(minzer);
                    }
                  else
                    {
                      VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] = VGNmax;
                      a1_dr_normal_int.point(ia, iy, iz) = (std::isnan(a1_init) ? -INFINITY : a1_init);
                    }
                  // Global search
                  double VGCmax = -INFINITY;
                  a1_init = NAN;
                  min_a2 = INFINITY;
                  max_a2 = -INFINITY;

                  for (size_t ia1 = 0; ia1 < ngrid_a1; ia1++)
                    {
                      double a1 = grid_a1[ia1];
                      double c = gdp(y, z)+a-g*a1*price(y, z, true, a1);
                      
                      if (c <= 0)
                        continue;
                      else
                        {
                          min_a2 = std::min(min_a2, a1);
                          max_a2 = std::max(max_a2, a1);
                        }

                      double VGCcur = util(c) + beta*pow(g, 1-gam)*EV1_bp[ia1];
                      if (VGCcur > VGCmax)
                        {
                          VGCmax = VGCcur;
                          a1_init = a1;
                        }
                    }

                  // Run the minimization
                  if (VGCmax != -INFINITY && (a1_init != min_a2 && a1_init != max_a2))
                    {
                      struct VG_params params = { a, y, z, quad_points_y1_bonus_possible, quad_points_z1_bonus_possible, this };
                      gsl_function F;
                      F.function = &CohenVillemotVFI::VGC_objective;
                      F.params = &params;

                      gsl_min_fminimizer_set(minzer, &F, a1_init, min_a2, max_a2);

                      int status, it2 = 0;
                      do
                        {
                          status = gsl_min_fminimizer_iterate(minzer);
                          assert(status != GSL_EBADFUNC && status != GSL_FAILURE);

                          double x1 = gsl_min_fminimizer_x_lower(minzer);
                          double x2 = gsl_min_fminimizer_x_upper(minzer);
     
                          status = gsl_min_test_interval(x1, x2, maximization_tol, 0.0);
                        }
                      while (status == GSL_CONTINUE && ++it2 < MAX_IT);
                      
                      VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] = -gsl_min_fminimizer_f_minimum(minzer);
                      a1_dr_bonus_possible_int.point(ia, iy, iz) = gsl_min_fminimizer_x_minimum(minzer);
                    }
                  else
                    {
                      VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] = VGCmax;
                      a1_dr_bonus_possible_int.point(ia, iy, iz) = (std::isnan(a1_init) ? -INFINITY : a1_init);
                    }
                }
            }
        gsl_min_fminimizer_free(minzer);
        delete[] quad_points_y1_normal;
        delete[] quad_points_z1_normal;
        delete[] quad_points_y1_bonus_possible;
        delete[] quad_points_z1_bonus_possible;
        delete[] quad_points_y1_VB;
        delete[] EV1_normal;
        delete[] EV1_bp;
      } // End #pragma parallel

      // Compute error
      err = 0;
      for (size_t iy = 0; iy < ngrid_y; iy++)
        for (size_t iz = 0; iz < ngrid_z; iz++)
          {
            err = std::max(err, fabs(VBN_points_new[iy+ngrid_y*iz]-VBNint.point(iy, iz)));
            err = std::max(err, fabs(VBC_points_new[iy+ngrid_y*iz]-VBCint.point(iy, iz)));
            for (size_t ia = 0; ia < ngrid_a; ia++)
              {
                if (VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] != -INFINITY
                    || VGNint.point(ia, iy, iz) != -INFINITY)
                  err = std::max(err, fabs(VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)]-VGNint.point(ia, iy, iz)));
                if ((VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] == -INFINITY
                     || VGNint.point(ia, iy, iz) == -INFINITY)
                    && VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] != VGNint.point(ia, iy, iz))
                  std::cout << "VGN at point a=" << grid_a[ia] << ", y=" << grid_y[iy]
                            << ", z=" << grid_z[iz]
                            << " switched to " << VGN_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] << std::endl;

                if (VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] != -INFINITY
                    || VGCint.point(ia, iy, iz) != -INFINITY)
                  err = std::max(err, fabs(VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)]-VGCint.point(ia, iy, iz)));
                if ((VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] == -INFINITY
                     || VGCint.point(ia, iy, iz) == -INFINITY)
                    && VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] != VGCint.point(ia, iy, iz))
                  std::cout << "VGC at point a=" << grid_a[ia] << ", y=" << grid_y[iy]
                            << ", z=" << grid_z[iz]
                            << " switched to " << VGC_points_new[ia+ngrid_a*(iy+ngrid_y*iz)] << std::endl;
              }
          }

      // Copy
      VBNint.init(VBN_points_new);
      VBCint.init(VBC_points_new);
      VGNint.init(VGN_points_new);
      VGCint.init(VGC_points_new);

      // Output
      std::cout << "Iteration " << it++ << ": error=" << err << std::endl;
    }
  while(err > convergence_tol && it < MAX_IT);

  delete[] VBN_points_new;
  delete[] VBC_points_new;
  delete[] VGN_points_new;
  delete[] VGC_points_new;

  a1_dr_normal_int.init();
  a1_dr_bonus_possible_int.init();
  
  if (it == MAX_IT)
    throw ConvergenceError("Maximum number of iterations reached");
}

double
CohenVillemotVFI::VGN_objective(double a1, void *params)
{
  struct VG_params *p = (struct VG_params *) params;
  const CohenVillemotVFI &CV = *p->CV;
  double y = p->y, a = p->a, z = p->z, g = exp(y)+z;

  double EV1 = 0;
  for (size_t i = 0; i < p->CV->quad_npoints; i++)
    EV1 += CV.quad_weights_normal[i] * CV.V(a1, p->quad_points_y1[i], p->quad_points_z1[i], CV.quad_points_bonus_possible_normal[i]);

  double v = -(CV.util(CV.gdp(y, z)+a-g*a1*CV.price(y, z, false, a1))
               + CV.beta*pow(g, 1-CV.gam)*EV1);

  if (std::isnan(v))
    return -almost_min_infinity;
  else
    return v;
}

double
CohenVillemotVFI::VGC_objective(double a1, void *params)
{
  struct VG_params *p = (struct VG_params *) params;
  const CohenVillemotVFI &CV = *p->CV;
  double y = p->y, a = p->a, z = p->z, g = exp(y)+z;

  double EV1 = 0;
  for (size_t i = 0; i < p->CV->quad_npoints; i++)
    EV1 += CV.quad_weights_bonus_possible[i] * CV.V(a1, p->quad_points_y1[i], p->quad_points_z1[i], CV.quad_points_bonus_possible_bonus_possible[i]);

  double v = -(CV.util(CV.gdp(y, z)+a-g*a1*CV.price(y, z, true, a1))
               + CV.beta*pow(g, 1-CV.gam)*EV1);

  if (std::isnan(v))
    return -almost_min_infinity;
  else
    return v;
}

double
CohenVillemotVFI::a1_dr(double a, double y, double z, bool bonus_possible) const
{
  assert(!std::isnan(y) && !std::isnan(z) && !std::isnan(a));
  return bonus_possible ? a1_dr_bonus_possible_int.get(a, y, z) : a1_dr_normal_int.get(a, y, z);
}
