/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cassert>

#include <algorithm>
#include <ext/algorithm>

#include <boost/lambda/lambda.hpp>

#include <gsl/gsl_math.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_poly.h>

#include "GaussQuadrature.hh"

GaussQuadrature::GaussQuadrature(double sigma_arg, size_t npoints_arg,
                                 gauss_t method_arg, double truncation_window_arg)
  : sigma(sigma_arg), npoints(npoints_arg), method(method_arg),
    truncation_window(truncation_window_arg)
{
  // Initialize quadrature points
  points = new double[npoints];
  weights = new double[npoints];
  switch (method)
    {
    case Legendre:
      {
        compute_legendre_points();
        // Adapt the quadrature points for a centered gaussian variable truncated from [-truncation_window*sigma, truncation_window*sigma]
        for (size_t i = 0; i < npoints; i++)
          weights[i] *= truncation_window/sqrt(M_PI)/M_SQRT2*exp(-gsl_pow_2(truncation_window*points[i])/2)/(1.0-2.0*gsl_cdf_gaussian_P(-truncation_window*sigma, sigma));
        std::transform(points, points+npoints, points, boost::lambda::_1*truncation_window*sigma);
      }
      break;
    case Hermite:
      compute_hermite_points();
      // Adapt the quadrature points for a standard gaussian variable
      std::transform(points, points+npoints, points, boost::lambda::_1*M_SQRT2);
      std::transform(weights, weights+npoints, weights, boost::lambda::_1/sqrt(M_PI));
      // Adapt for the variance
      std::transform(points, points+npoints, points, boost::lambda::_1*sigma);
      break;
    }
}

GaussQuadrature::GaussQuadrature(const GaussQuadrature &arg)
  : sigma(arg.sigma), npoints(arg.npoints), method(arg.method),
    truncation_window(arg.truncation_window)
{  
  points = new double[npoints];
  __gnu_cxx::copy_n(arg.points, npoints, points);
  weights = new double[npoints];
  __gnu_cxx::copy_n(arg.weights, npoints, weights);
}

GaussQuadrature::~GaussQuadrature()
{
  delete[] points;
  delete[] weights;
}

void
GaussQuadrature::compute_hermite_points()
{
  assert(npoints >= 1);
  double *P0, *P1, *P_tmp;
  P0 = new double[npoints+1];
  P1 = new double[npoints+1];
  P_tmp = new double[npoints+1];

  P0[0] = 1.0;
  std::fill_n(P0+1, npoints, 0.0);

  P1[0] = 0.0;
  P1[1] = 2.0;
  std::fill_n(P1+2, npoints-1, 0.0);

  for (size_t i = 1; i < npoints; i++)
    {
      // Computes P_tmp = 2*x*P1-2*i*P0
      P_tmp[0] = 0.0;
      std::transform(P1, P1+npoints, P_tmp+1, 2.0*boost::lambda::_1);
      std::transform(P_tmp, P_tmp+npoints+1, P0, P_tmp,
                     (boost::lambda::_1 - 2*((double) i)*boost::lambda::_2));
      std::copy(P1, P1+npoints+1, P0);
      std::copy(P_tmp, P_tmp+npoints+1, P1);
    }

  gsl_poly_complex_workspace *w = gsl_poly_complex_workspace_alloc(npoints+1);
  double *roots = new double[2*npoints];
  gsl_poly_complex_solve(P1, npoints+1, w, roots);

  for (size_t i = 0; i < npoints; i++)
    {
      points[i] = roots[2*i];
      weights[i] = gsl_pow_int(2.0, npoints-1)*factorial(npoints)*sqrt(M_PI)/gsl_pow_2(npoints)/gsl_pow_2(gsl_poly_eval(P0, npoints+1, points[i]));
    }

  delete[] P0;
  delete[] P1;
  delete[] P_tmp;
  delete[] roots;
  gsl_poly_complex_workspace_free(w);
}

void
GaussQuadrature::compute_legendre_points()
{
  assert(npoints >= 1);
  double *P0, *P1, *P_tmp;
  P0 = new double[npoints+1];
  P1 = new double[npoints+1];
  P_tmp = new double[npoints+1];

  P0[0] = 1.0;
  std::fill_n(P0+1, npoints, 0.0);

  P1[0] = 0.0;
  P1[1] = 1.0;
  std::fill_n(P1+2, npoints-1, 0.0);

  for (size_t i = 1; i < npoints; i++)
    {
      // Computes P_tmp = ((2*i+1)*x*P1-i*P0)/(i+1)
      P_tmp[0] = 0.0;
      std::transform(P1, P1+npoints, P_tmp+1, (2.0*((double) i)+1.0)*boost::lambda::_1);
      std::transform(P_tmp, P_tmp+npoints+1, P0, P_tmp,
                     (boost::lambda::_1 - ((double) i)*boost::lambda::_2));
      std::transform(P_tmp, P_tmp+npoints+1, P_tmp, boost::lambda::_1/(((double) i) + 1.0));
      std::copy(P1, P1+npoints+1, P0);
      std::copy(P_tmp, P_tmp+npoints+1, P1);
    }

  gsl_poly_complex_workspace *w = gsl_poly_complex_workspace_alloc(npoints+1);
  double *roots = new double[2*npoints];
  gsl_poly_complex_solve(P1, npoints+1, w, roots);

  for (size_t i = 0; i < npoints; i++)
    {
      points[i] = roots[2*i];
      double res[2];
      gsl_poly_eval_derivs(P1, npoints+1, points[i], res, 2);
      weights[i] = 2.0/(1.0-gsl_pow_2(points[i]))/gsl_pow_2(res[1]);
    }

  delete[] P0;
  delete[] P1;
  delete[] P_tmp;
  delete[] roots;
  gsl_poly_complex_workspace_free(w);
}

double
GaussQuadrature::factorial(size_t n) const
{
  if (n == 1)
    return 1;
  else
    return ((double) n)*factorial(n-1);
}
