/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cassert>

#include <gsl/gsl_math.h>

#include <ext/algorithm>

#include "Interpolator1D.hh"

Interpolator1D::Interpolator1D(size_t ngrid_arg, const gsl_interp_type *type_arg)
  : ngrid(ngrid_arg), type(type_arg)
{
  grid = new double[ngrid];
  
  points = new double[ngrid];
  spline = gsl_interp_alloc(type, ngrid);
}

Interpolator1D::Interpolator1D(const Interpolator1D &arg)
  : ngrid(arg.ngrid), type(arg.type)
{
  grid = new double[ngrid];
  __gnu_cxx::copy_n(arg.grid, ngrid, grid);

  points = new double[ngrid];
  __gnu_cxx::copy_n(arg.points, ngrid, points);
  
  spline = gsl_interp_alloc(type, ngrid);

  init();
}

Interpolator1D::~Interpolator1D()
{
  delete[] grid;
  delete[] points;
  gsl_interp_free(spline);
}

void
Interpolator1D::set_grid(const double *grid_arg)
{
  __gnu_cxx::copy_n(grid_arg, ngrid, grid);
}

void
Interpolator1D::init()
{
  gsl_interp_init(spline, grid, points, ngrid);
}

void
Interpolator1D::init(const double *points_arg)
{
  __gnu_cxx::copy_n(points_arg, ngrid, points);
  init();
}

double &
Interpolator1D::point(size_t i)
{
  return points[i];
}

double
Interpolator1D::get(double x) const
{
  assert(!std::isnan(x));

  double x1 = std::min(std::max(x, grid[0]), grid[ngrid-1]);

  double v = gsl_interp_eval(spline, grid, points, x1, NULL);

  if (x != x1)
    v += gsl_interp_eval_deriv(spline, grid, points, x1, NULL) * (x-x1);

  return v;
}
