/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cmath>
#include <cassert>

#include <gsl/gsl_math.h>

#include <ext/algorithm>

#include "Interpolator3D.hh"

Interpolator3D::Interpolator3D(size_t ngrid1_arg, const gsl_interp_type *type1_arg,
                               size_t ngrid2_arg, const gsl_interp_type *type2_arg,
                               size_t ngrid3_arg, const gsl_interp_type *type3_arg,
                               bool tensor_product_grid_arg)
  : ngrid1(ngrid1_arg), type1(type1_arg),
    ngrid2(ngrid2_arg), type2(type2_arg),
    ngrid3(ngrid3_arg), type3(type3_arg),
    tensor_product_grid(tensor_product_grid_arg)
{
  if (tensor_product_grid)
    grid1 = new double[ngrid1];
  else
    grid1 = new double[ngrid1*ngrid2*ngrid3];
  grid2 = new double[ngrid2];
  grid3 = new double[ngrid3];
  
  points = new double[ngrid1*ngrid2*ngrid3];
  spline = (gsl_interp **) malloc(sizeof(gsl_interp *)*ngrid2*ngrid3);
  shifts = new size_t[ngrid2*ngrid3];
  for (size_t i2 = 0; i2 < ngrid2; i2++)
    for (size_t i3 = 0; i3 < ngrid3; i3++)
      {
        spline[i2+ngrid2*i3] = gsl_interp_alloc(type1, ngrid1);
        shifts[i2+ngrid2*i3] = 0;
      }
}

Interpolator3D::Interpolator3D(const Interpolator3D &arg)
  : ngrid1(arg.ngrid1), type1(arg.type1),
    ngrid2(arg.ngrid2), type2(arg.type2),
    ngrid3(arg.ngrid3), type3(arg.type3),
    tensor_product_grid(arg.tensor_product_grid)
{
  if (tensor_product_grid)
    {
      grid1 = new double[ngrid1];
      __gnu_cxx::copy_n(arg.grid1, ngrid1, grid1);
    }
  else
    {
      grid1 = new double[ngrid1*ngrid2*ngrid3];
      __gnu_cxx::copy_n(arg.grid1, ngrid1*ngrid2*ngrid3, grid1);
    }
  grid2 = new double[ngrid2];
  __gnu_cxx::copy_n(arg.grid2, ngrid2, grid2);
  grid3 = new double[ngrid3];
  __gnu_cxx::copy_n(arg.grid3, ngrid3, grid3);

  points = new double[ngrid1*ngrid2*ngrid3];
  __gnu_cxx::copy_n(arg.points, ngrid1*ngrid2*ngrid3, points);
  
  spline = (gsl_interp **) malloc(sizeof(gsl_interp *)*ngrid2*ngrid3);
  shifts = new size_t[ngrid2*ngrid3];
  for (size_t i2 = 0; i2 < ngrid2; i2++)
    for (size_t i3 = 0; i3 < ngrid3; i3++)
      {
        shifts[i2+ngrid2*i3] = arg.shifts[i2+ngrid2*i3];
        spline[i2+ngrid2*i3] = gsl_interp_alloc(type1, ngrid1 - arg.shifts[i2+ngrid2*i3]);
      }

  init();
}

Interpolator3D::~Interpolator3D()
{
  delete[] grid1;
  delete[] grid2;
  delete[] grid3;
  
  delete[] points;
  for (size_t i = 0; i < ngrid2*ngrid3; i++)
    gsl_interp_free(spline[i]);
  free(spline);
  delete[] shifts;
}

void
Interpolator3D::set_grids(const double *grid1_arg, const double *grid2_arg, const double *grid3_arg)
{
  if (tensor_product_grid)
    __gnu_cxx::copy_n(grid1_arg, ngrid1, grid1);
  else
    __gnu_cxx::copy_n(grid1_arg, ngrid1*ngrid2*ngrid3, grid1);
  __gnu_cxx::copy_n(grid2_arg, ngrid2, grid2);
  __gnu_cxx::copy_n(grid3_arg, ngrid3, grid3);
}

void
Interpolator3D::init()
{
  for (size_t i2 = 0; i2 < ngrid2; i2++)
    for (size_t i3 = 0; i3 < ngrid3; i3++)
      {
        size_t s = ngrid1;
        while(s > 0 && points[ngrid1*(i2+ngrid2*i3)+s-1] > -INFINITY)
          s--;

        if (s != shifts[i2+ngrid2*i3])
          {
            if (spline[i2+ngrid2*i3] != NULL)
              free(spline[i2+ngrid2*i3]);
            if (ngrid1 - s > gsl_interp_type_min_size(type1))
              spline[i2+ngrid2*i3] = gsl_interp_alloc(type1, ngrid1 - s);
            else if (ngrid1 - s > gsl_interp_type_min_size(gsl_interp_linear))
              spline[i2+ngrid2*i3] = gsl_interp_alloc(gsl_interp_linear, ngrid1 - s);
            else
              spline[i2+ngrid2*i3] = NULL;
            shifts[i2+ngrid2*i3] = s;
          }

        size_t tensor_offset = tensor_product_grid ? 0 : ngrid1*(i2+ngrid2*i3);
        if (spline[i2+ngrid2*i3])
          gsl_interp_init(spline[i2+ngrid2*i3], grid1 + tensor_offset + s, points + ngrid1*(i2+ngrid2*i3) + s, ngrid1 - s);
      }
}

void
Interpolator3D::init(const double *points_arg)
{
  __gnu_cxx::copy_n(points_arg, ngrid1*ngrid2*ngrid3, points);
  init();
}

double &
Interpolator3D::point(size_t i1, size_t i2, size_t i3)
{
  return points[i1 + ngrid1*(i2+ngrid2*i3)];
}

double
Interpolator3D::get(double x, double y, double z) const
{
  assert(!std::isnan(x) && !std::isnan(y) && !std::isnan(z));

  double y1 = std::min(std::max(y, grid2[0]), grid2[ngrid2-1]);
  double z1 = std::min(std::max(z, grid3[0]), grid3[ngrid3-1]);

  double *points_tmp = new double[ngrid2*ngrid3];
  double *points_tmp2 = new double[ngrid3];
  size_t *shifts_tmp = new size_t[ngrid3];
  
  std::fill_n(shifts_tmp, ngrid3, 0);
  
  for (size_t i2 = 0; i2 < ngrid2; i2++)
    for (size_t i3 = 0; i3 < ngrid3; i3++)
      {
        size_t i = i2+ngrid2*i3;
        size_t tensor_offset = tensor_product_grid ? 0 : ngrid1*i;
        if (spline[i] == NULL || (shifts[i] > 0 && x < grid1[tensor_offset + shifts[i]]))
          shifts_tmp[i3] = std::max(i2+1, shifts_tmp[i3]);
        else
          {
            double x1 = std::min(std::max(x, grid1[tensor_offset]), grid1[tensor_offset+ngrid1-1]);

            points_tmp[i] = gsl_interp_eval(spline[i], grid1 + tensor_offset + shifts[i], points + ngrid1*i + shifts[i], x1, NULL);

            if (x1 != x)
              points_tmp[i] += gsl_interp_eval_deriv(spline[i], grid1 + tensor_offset + shifts[i], points + i*ngrid1 + shifts[i], x1, NULL)*(x-x1);
          }
      }

  size_t shift_tmp2 = 0;
  
  for (size_t i3 = 0; i3 < ngrid3; i3++)
    if (ngrid2-shifts_tmp[i3] < std::min(gsl_interp_type_min_size(type2),
                                         gsl_interp_type_min_size(gsl_interp_linear))
        || (shifts_tmp[i3] > 0 && y1 < grid2[shifts_tmp[i3]]))
      shift_tmp2 = std::max(i3+1, shift_tmp2);
    else
      {
        gsl_interp *spline_tmp;
        if (ngrid2-shifts_tmp[i3] >= gsl_interp_type_min_size(type2))
          spline_tmp = gsl_interp_alloc(type2, ngrid2-shifts_tmp[i3]);
        else
          spline_tmp = gsl_interp_alloc(gsl_interp_linear, ngrid2-shifts_tmp[i3]);
        gsl_interp_init(spline_tmp, grid2 + shifts_tmp[i3], points_tmp + ngrid2*i3 + shifts_tmp[i3], ngrid2 - shifts_tmp[i3]);
        points_tmp2[i3] = gsl_interp_eval(spline_tmp, grid2 + shifts_tmp[i3], points_tmp + ngrid2*i3 + shifts_tmp[i3], y1, NULL);

        if (y != y1)
          points_tmp2[i3] += gsl_interp_eval_deriv(spline_tmp, grid2 + shifts_tmp[i3], points_tmp + ngrid2*i3 + shifts_tmp[i3], y1, NULL)*(y-y1);
        
        gsl_interp_free(spline_tmp);
      }

  if (ngrid3-shift_tmp2 < std::min(gsl_interp_type_min_size(type3),
                                   gsl_interp_type_min_size(gsl_interp_linear))
      || (shift_tmp2 > 0 && z1 < grid3[shift_tmp2]))
    {
      delete[] points_tmp;
      delete[] points_tmp2;
      delete[] shifts_tmp;

      return -INFINITY;
    }
  
  gsl_interp *spline_tmp;
  if (ngrid3-shift_tmp2 >= gsl_interp_type_min_size(type3))
    spline_tmp = gsl_interp_alloc(type3, ngrid3-shift_tmp2);
  else
   spline_tmp = gsl_interp_alloc(gsl_interp_linear, ngrid3-shift_tmp2);
  gsl_interp_init(spline_tmp, grid3 + shift_tmp2, points_tmp2 + shift_tmp2, ngrid3 - shift_tmp2);
  double v = gsl_interp_eval(spline_tmp, grid3 + shift_tmp2, points_tmp2 + shift_tmp2, z1, NULL);

  if (z != z1)
    v += gsl_interp_eval_deriv(spline_tmp, grid3 + shift_tmp2, points_tmp2 + shift_tmp2, z1, NULL)*(z-z1);
  
  gsl_interp_free(spline_tmp);
  delete[] points_tmp;
  delete[] points_tmp2;
  delete[] shifts_tmp;
  
  return v;
}
