/*
 * Copyright (C) 2012 Sébastien Villemot <sebastien@dynare.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef INTERPOLATOR_3D_HH
#define INTERPOLATOR_3D_HH

#include <gsl/gsl_interp.h>

//! Interpolates a function of three variables
/*! 
  Also contains a mechanism to deal with the case where the function is equal
  to minus infinity in some areas of its domain: the interpolation will still
  work (potentially returning minus infinity).

  If the value of the function is minus infinity in some of its grid points,
  the code assumes that the function is increasing in all its arguments.

  Note that this class is able to interpolate outside the grid domain, using
  a first order Taylor-expansion outside the grid.
  
  If tensor_product_grid=false, then the class is able to deal with a grid
  which is not a tensor-product grid. The grid is still of size ngrid1*ngrid2*ngrid3,
  but for each couple in grid2*grid3, there is a different grid over the first
  dimension. grid2 and grid3 are therefore still vector of sizes ngrid2 and
  ngrid3 respectively, but grid1 is a 3D array of size ngrid1*ngrid2*ngrid3.
*/
class Interpolator3D
{
public:
  //! Main constructor
  /*! 
    \param ngrid1_arg Size of grid in first dimension
    \param type1_arg Type of interpolation in first dimension
    \param ngrid2_arg Size of grid in second dimension
    \param type2_arg Type of interpolation is second dimension
    \param ngrid3_arg Size of grid in third dimension
    \param type3_arg Type of interpolation is third dimension
    \param tensor_product_grid_arg Whether the grid is tensor product
  */
  Interpolator3D(size_t ngrid1_arg, const gsl_interp_type *type1_arg,
                 size_t ngrid2_arg, const gsl_interp_type *type2_arg,
                 size_t ngrid3_arg, const gsl_interp_type *type3_arg,
                 bool tensor_product_grid_arg = true);
  //! Copy constructor
  Interpolator3D(const Interpolator3D &arg);
  //! Destructor
  ~Interpolator3D();

  const size_t ngrid1;          //!< Size of grid in first dimension
  const gsl_interp_type *type1; //!< Type of interpolation in first dimension

  const size_t ngrid2;          //!< Size of grid in second dimension
  const gsl_interp_type *type2; //!< Type of interpolation in second dimension

  const size_t ngrid3;          //!< Size of grid in third dimension
  const gsl_interp_type *type3; //!< Type of interpolation in third dimension

  const bool tensor_product_grid; //!< Whether the grid is tensor product

  //! Set the interpolation grid
  /*! 
    \param grid1_arg Grid points in first dimension
    \param grid2_arg Grid points in second dimension
    \param grid3_arg Grid points in third dimension
  */
  void set_grids(const double *grid1_arg, const double *grid2_arg, const double *grid3_arg);

  //! Inits the internal interpolation engine and compute the shifts
  /*!
    Has to be called after interpolation values have been modified through points()
  */
  void init();

  //! Set the value of the function at all grid points
  /*! 
    It also initializes the interpolation engine and computes the shifts
    \param points_arg Value of the function at grid points (first dim has major order)
  */
  void init(const double *points_arg);

  //! Get or set the value of the function at a given grid point
  /*! 
    It is necessary to call init() after this function for proper interpolation computations.
    \param i1 The first dimension index of the grid point
    \param i2 The second dimension index of the grid point
    \param i3 The third dimension index of the grid point
    
    \return An l-value pointing to the value at the requested point
  */
  double &point(size_t i1, size_t i2, size_t i3);

  //! Compute the interpolated value at an arbitrary point
  /*! 
    \param x First coordinate of requested point
    \param y Second coordinate of requested point
    \param z Third coordinate of requested point
    
    \return The interpolated value
  */
  double get(double x, double y, double z) const;
private:
  double *grid1,                //!< Grid points in first dimension
    *grid2,                     //!< Grid points in second dimension
    *grid3;                     //!< Grid points in third dimension
  double *points;               //!< Value at grid points (first dim has major order)
  gsl_interp **spline;          //!< 2D array of interpolation engines

  //! Shift index for dealing with -inf values
  /*! 
    shifts[i2, i3] = minimum { i1 such that point(i1, i2, i3) > -inf }

    The array is stored with i2 in major order
   */
  size_t *shifts;
};

#endif
